static char *cvs_id =" $Id: 7c_use.c,v 1.1.1.1 2002/01/03 14:52:03 berar Exp $ ";
/*   7c_lib.c     functions to be used on d2am 7circles goniometer  */

char *date_7c_use = __DATE__  ;
char *time_7c_use = __TIME__  ;

#include "7c_def.h"

/* ------------------------------------------------------*/
#define DEBUG(order)          /* order  */
#define ROT_DEBUG(order)      /* order  */
/* ------------------------------------------------------*/
#define EPSILON   1.0e-20
#define EPS_ERR   2.0e-5
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/*     Following functions are for internal use          */
/*        simple math common functions                   */
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/*   return the number of solution for the equation  :   */
/*       a cos(phi) + b sin(phi) + c = 0                 */
/*   the array sol is filed with the angle in degree     */
/* ------------------------------------------------------*/
int solve_trigo ( double *sol, double a, double b, double c)
{
    int i, m, n;
    double s_phi, c_phi, phi;
    double loc_sol[2];
    double ea, eb;
    DEBUG(printf("solve_trigo a=%lg  b=%lg  c=%lg\n", a, b, c);)
    if ( (fabs(a)<EPSILON) &&  (fabs(b)<EPSILON) && (fabs(c)<5.0e-7))
    {   sol[0] = 0;                 /* 0 = 0 */
        return 1;
        }   
    if ( fabs(a) > fabs(b) )        /* solve using cosinus */
    {    if ( fabs(b) < EPSILON )   /* a cos(phi) + c = 0 */
         {    if ( ( fabs(a) < EPSILON ) && ( fabs(c) < EPSILON ) )
              {   sol[0] = 0;                 /* 0 = 0 */
                  return 1;
                  }   
              c_phi = -c / a;
              if (fabs(c_phi) > 1) return 0;
              sol[0] = phi = acos(c_phi);
              sol[1] = -phi;
              return 2;
              }
          n=solve_x2( loc_sol, a*a+b*b, 2*a*c, c*c-b*b);
          if (n == 0 )return 0;
          for ( i = m = 0; i < n; i++)
          {   if ( fabs(loc_sol[i]) > 1 ) break;
              phi = acos(loc_sol[i]);
              s_phi = sin(phi);
              ea = a * loc_sol[i] + b * s_phi + c;
              eb = a * loc_sol[i] - b * s_phi + c;
              if ( fabs(ea) > fabs(eb) ) phi = -phi;
              sol[m++] = phi;
              }
          }    
    else      /* solve  using sinus */
    {    if ( fabs(a) < EPSILON )   /* b sin(phi) + c = 0 */
         {    if ( ( fabs(b) < EPSILON ) && ( fabs(c) < EPSILON ) )
              {   sol[0] = 0;                 /* 0 = 0 */
                  return 1;
                  }   
              s_phi = -c / b;
              if (fabs(s_phi) > 1) return 0;
              sol[0] = phi = asin(s_phi);
              sol[1] = M_PI-phi;
              return 2;
              }
          n = solve_x2( loc_sol, a*a+b*b, 2*b*c, c*c-a*a); 
          if (n == 0 )return 0;
          for ( i = m = 0; i < n; i++)
          {   if ( fabs(loc_sol[i]) > 1 ) break;
              phi = asin(loc_sol[i]);
              c_phi = cos(phi);
              ea = b * loc_sol[i] + a * c_phi + + c;
              eb = b * loc_sol[i] - a * c_phi + c;
              if ( fabs(ea) > fabs(eb) ) phi = M_PI-phi;
              sol[m++] = phi;
              }
          }
     return m;
     }         
/* ------------------------------------------------------*/
/*   return the number of solution for the equation  :   */
/*       a x ^2 + b x + c = 0                           */
/*   the array sol is filed with the solutions           */
/* ------------------------------------------------------*/
int solve_x2( double *sol, double a, double b, double c)          
{             
    double s, dd;
    if ( fabs(a) < EPSILON )
    {  if ( fabs(b) < EPSILON ) return 0;
       sol[0] = - c / b;
       return 1;
       }
    dd = b * b - 4 * a *c;
    if ( dd < 0 ) return 0;
    s = - 0.5 * b / a;
    if ( dd < EPSILON )
    {    sol[0] = s;
         return 1;
         }
    dd = 0.5 * sqrt(dd) / a;
    sol[0] = s + dd;
    sol[1] = s - dd;
    return 2;
    }     
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/*   return the number of solution for the equation  :   */
/*       rot u = w     axe = 0 (x), 1 (y), 2 (z)         */
/*   the array sol is filed with the solutions           */
/* ------------------------------------------------------*/
int solve_rot( int axe, double * sol, vector_type *u, vector_type *w)
{   int x1, x2;
    double d, e, s, c;
    if(fabs(u->v[axe] - w->v[axe]) > EPS_ERR ) return 0;
    x1=(axe+1)%3;
    x2=(x1+1)%3;
    d = u->v[x1] * u->v[x1] + u->v[x2] * u->v[x2];
    e = w->v[x1] * w->v[x1] + w->v[x2] * w->v[x2];
    if ( d - e > EPS_ERR ) return 0;
    if ( d < EPS_ERR )
    {   *sol = 0;
        return 1;
        }
/*   printf( "solve_rot u %lf %lf w %lf %lf\n", u->v[x1], u->v[x2], w->v[x1], w->v[x2]);*/
   c  = w->v[x1] * u->v[x1] + w->v[x2] * u->v[x2]; 
   s  = u->v[x2] * w->v[x1] - u->v[x1] * w->v[x2]; 
   if ( fabs(c) - d > EPS_ERR ) return 0;
   if ( fabs(s) - d > EPS_ERR ) return 0;
   if ( fabs(c) > fabs(s) ) 
   {   *sol = asin(s/d);
       if ( c < 0 ) *sol = M_PI - *sol;
       }
   else
   {   *sol = acos(c/d);
       if ( s < 0 ) *sol = - *sol;
       }
/*   printf( "solve_rot %c  %lf ( c=%lf, s=%lf ,d=%lf)\n", 'x'+axe, *sol*TO_DEG, c, s, d );*/
   return 1;
   }    
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/*   rotation bloc                                       */
/* ------------------------------------------------------*/
void rotate ( int axe, vector_type *r, double angle, vector_type *s)
{   double sa, ca;
    int x1, x2;
    sa = sin(angle);
    ca = cos(angle);
    x1 = (axe+1) % 3;
    x2 = (x1+1) % 3;
    r->v[axe] = s->v[axe];
    r->v[x1] = ca * s->v[x1] + sa * s->v[x2];
    r->v[x2] = -sa * s->v[x1] + ca * s->v[x2];
    ROT_DEBUG(printf("ROT %c %7.3lg (%9.3lg %9.3lg %9.3lg)->(%9.3lg %9.3lg %9.3lg)\n", 'X'+axe, angle*TO_DEG, s->coord.x, s->coord.y, s->coord.z, r->coord.x, r->coord.y, r->coord.z);)
    }
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/*    3*3 matrix operations                              */
/* ------------------------------------------------------*/
void mul_matvec(vector_type *res, matrix_type *mat, vector_type *vec )
{   int i, j;
    for ( i = 0; i < 3; i++ ) 
    {    res->v[i] = 0;
         for ( j = 0; j < 3; j++ ) res->v[i] += mat->m[i][j]*vec->v[j];
         } 
    }
/* ------------------------------------------------------*/
void mul_matmat(matrix_type *res, matrix_type *mat, matrix_type *vec)
{   int i;
    for ( i = 0; i < 3; i++ ) mul_matvec( &res->v[i], mat, &vec->v[i]);
    }
/* ------------------------------------------------------*/
int inv_mat(matrix_type *res, matrix_type *mat )
{   int i, j, k;
    int l, m, n;
    double det = 0;
    for ( i = 0; i < 3; i++ )
    {   j = (i+1) % 3;
        k = (j+1) % 3;
        det += mat->m[i][0] * (mat->m[j][1] * mat->m[k][2] - mat->m[k][1] * mat->m[j][2]);
        }
    if ( fabs(det) <= 1.0e-40 ) 
    {   fprintf( stderr," Failed inverting 3*3 matrix!\n");
        return 1;
        }
    for ( i = 0; i < 3; i++ )
    {   j = (i+1) % 3;  k = (j+1) % 3;
        for ( l = 0; l < 3; l++ )
        {   m = (l+1) % 3;  n = (m+1) % 3;
            res->m[l][i] = ( mat->m[j][m] * mat->m[k][n] - mat->m[k][m] * mat->m[j][n]) / det;
            }
         }    
    return 0;            
    }
/* ------------------------------------------------------*/
/* ------------------------------------------------------*/
/*   print used for debug,....                           */
/* ------------------------------------------------------*/
void put_vector( char * id, vector_type * v)
{   printf(" %10s    %9.3lg   %9.3lg  %9.3lg\n", id, v->coord.x, v->coord.y, v->coord.z);
    }
/* ------------------------------------------------------*/
void put_matrix( char * id, matrix_type * m)
{   int i, j ;
    for ( i = 0; i < 3; i++) 
    {    printf(" %15s ", i==0?id:" ");
         for (j = 0; j < 3; j++) 
             printf(" [%d,%d]=%12.6lf %s", i, j, m->m[i][j], j==2?"\n":" ");
         }
    } 
/* ------------------------------------------------------*/
     
